SQLite — SQLAlchemy 1.4 Documentation

Release: 1.4.44 | Release Date: November 12, 2022

# SQLAlchemy 1.4 Documentation

# SQLite

Support for the SQLite database.

The following table summarizes current support levels for database release versions.

Supported SQLite versions
Support type | Versions
--- | ---
Fully tested in CI | 3.21, 3.28+
Normal support | 3.12+
Best effort | 3.7.16+

## DBAPI Support

The following dialect/DBAPI options are available. Please refer to individual DBAPI sections for connect information.

- pysqlite

- aiosqlite

- pysqlcipher

## Date and Time Types

SQLite does not have built-in DATE, TIME, or DATETIME types, and pysqlite does not provide out of the box functionality for translating values between Python datetime objects and a SQLite-supported format. SQLAlchemy’s own `DateTime` and related types provide date formatting and parsing functionality when SQLite is used. The implementation classes are `DATETIME`, `DATE` and `TIME`. These types represent dates and times as ISO formatted strings, which also nicely support ordering. There’s no reliance on typical “libc” internals for these functions so historical dates are fully supported.

### Ensuring Text affinity

The DDL rendered for these types is the standard `DATE`, `TIME` and `DATETIME` indicators. However, custom storage formats can also be applied to these types. When the storage format is detected as containing no alpha characters, the DDL for these types is rendered as `DATE_CHAR`, `TIME_CHAR`, and `DATETIME_CHAR`, so that the column continues to have textual affinity.

See also

Type Affinity - in the SQLite documentation

## SQLite Auto Incrementing Behavior

Background on SQLite’s autoincrement is at: https://sqlite.org/autoinc.html

Key concepts:

- SQLite has an implicit “auto increment” feature that takes place for any non-composite primary-key column that is specifically created using “INTEGER PRIMARY KEY” for the type + primary key.

- SQLite also has an explicit “AUTOINCREMENT” keyword, that is not equivalent to the implicit autoincrement feature; this keyword is not recommended for general use. SQLAlchemy does not render this keyword unless a special SQLite-specific directive is used (see below). However, it still requires that the column’s type is named “INTEGER”.

### Using the AUTOINCREMENT Keyword

To specifically render the AUTOINCREMENT keyword on the primary key column when rendering DDL, add the flag `sqlite_autoincrement=True` to the Table construct:

    Table('sometable', metadata,
            Column('id', Integer, primary_key=True),
            sqlite_autoincrement=True)

### Allowing autoincrement behavior SQLAlchemy types other than Integer/INTEGER

SQLite’s typing model is based on naming conventions. Among other things, this means that any type name which contains the substring `"INT"` will be determined to be of “integer affinity”. A type named `"BIGINT"`, `"SPECIAL_INT"` or even `"XYZINTQPR"`, will be considered by SQLite to be of “integer” affinity. However, the SQLite autoincrement feature, whether implicitly or explicitly enabled, requires that the name of the column’s type is exactly the string “INTEGER”. Therefore, if an application uses a type like `BigInteger` for a primary key, on SQLite this type will need to be rendered as the name `"INTEGER"` when emitting the initial `CREATE TABLE` statement in order for the autoincrement behavior to be available.

One approach to achieve this is to use `Integer` on SQLite only using `TypeEngine.with_variant()`:

    table = Table(
        "my_table", metadata,
        Column("id", BigInteger().with_variant(Integer, "sqlite"), primary_key=True)
    )

Another is to use a subclass of `BigInteger` that overrides its DDL name to be `INTEGER` when compiled against SQLite:

    from sqlalchemy import BigInteger
    from sqlalchemy.ext.compiler import compiles

    class SLBigInteger(BigInteger):
        pass

    @compiles(SLBigInteger, 'sqlite')
    def bi_c(element, compiler, **kw):
        return "INTEGER"

    @compiles(SLBigInteger)
    def bi_c(element, compiler, **kw):
        return compiler.visit_BIGINT(element, **kw)

    table = Table(
        "my_table", metadata,
        Column("id", SLBigInteger(), primary_key=True)
    )

See also

`TypeEngine.with_variant()`

Custom SQL Constructs and Compilation Extension

Datatypes In SQLite Version 3

## Database Locking Behavior / Concurrency

SQLite is not designed for a high level of write concurrency. The database itself, being a file, is locked completely during write operations within transactions, meaning exactly one “connection” (in reality a file handle) has exclusive access to the database during this period - all other “connections” will be blocked during this time.

The Python DBAPI specification also calls for a connection model that is always in a transaction; there is no `connection.begin()` method, only `connection.commit()` and `connection.rollback()`, upon which a new transaction is to be begun immediately. This may seem to imply that the SQLite driver would in theory allow only a single filehandle on a particular database file at any time; however, there are several factors both within SQLite itself as well as within the pysqlite driver which loosen this restriction significantly.

However, no matter what locking modes are used, SQLite will still always lock the database file once a transaction is started and DML (e.g. INSERT, UPDATE, DELETE) has at least been emitted, and this will block other transactions at least at the point that they also attempt to emit DML. By default, the length of time on this block is very short before it times out with an error.

This behavior becomes more critical when used in conjunction with the SQLAlchemy ORM. SQLAlchemy’s `Session` object by default runs within a transaction, and with its autoflush model, may emit DML preceding any SELECT statement. This may lead to a SQLite database that locks more quickly than is expected. The locking mode of SQLite and the pysqlite driver can be manipulated to some degree, however it should be noted that achieving a high degree of write-concurrency with SQLite is a losing battle.

For more information on SQLite’s lack of write concurrency by design, please see Situations Where Another RDBMS May Work Better - High Concurrency near the bottom of the page.

The following subsections introduce areas that are impacted by SQLite’s file-based architecture and additionally will usually require workarounds to work when using the pysqlite driver.

## Transaction Isolation Level / Autocommit

SQLite supports “transaction isolation” in a non-standard way, along two axes. One is that of the PRAGMA read_uncommitted instruction. This setting can essentially switch SQLite between its default mode of `SERIALIZABLE` isolation, and a “dirty read” isolation mode normally referred to as `READ UNCOMMITTED`.

SQLAlchemy ties into this PRAGMA statement using the `create_engine.isolation_level` parameter of `create_engine()`. Valid values for this parameter when used with SQLite are `"SERIALIZABLE"` and `"READ UNCOMMITTED"` corresponding to a value of 0 and 1, respectively. SQLite defaults to `SERIALIZABLE`, however its behavior is impacted by the pysqlite driver’s default behavior.

When using the pysqlite driver, the `"AUTOCOMMIT"` isolation level is also available, which will alter the pysqlite connection using the `.isolation_level` attribute on the DBAPI connection and set it to None for the duration of the setting.

New in version 1.3.16: added support for SQLite AUTOCOMMIT isolation level when using the pysqlite / sqlite3 SQLite driver.

The other axis along which SQLite’s transactional locking is impacted is via the nature of the `BEGIN` statement used. The three varieties are “deferred”, “immediate”, and “exclusive”, as described at BEGIN TRANSACTION. A straight `BEGIN` statement uses the “deferred” mode, where the database file is not locked until the first read or write operation, and read access remains open to other transactions until the first write operation. But again, it is critical to note that the pysqlite driver interferes with this behavior by not even emitting BEGIN until the first write operation.

Warning

SQLite’s transactional scope is impacted by unresolved issues in the pysqlite driver, which defers BEGIN statements to a greater degree than is often feasible. See the section Serializable isolation / Savepoints / Transactional DDL for techniques to work around this behavior.

See also

Setting Transaction Isolation Levels including DBAPI Autocommit

## SAVEPOINT Support

SQLite supports SAVEPOINTs, which only function once a transaction is begun. SQLAlchemy’s SAVEPOINT support is available using the `Connection.begin_nested()` method at the Core level, and `Session.begin_nested()` at the ORM level. However, SAVEPOINTs won’t work at all with pysqlite unless workarounds are taken.

Warning

SQLite’s SAVEPOINT feature is impacted by unresolved issues in the pysqlite driver, which defers BEGIN statements to a greater degree than is often feasible. See the section Serializable isolation / Savepoints / Transactional DDL for techniques to work around this behavior.

## Transactional DDL

The SQLite database supports transactional DDL as well. In this case, the pysqlite driver is not only failing to start transactions, it also is ending any existing transaction when DDL is detected, so again, workarounds are required.

Warning

SQLite’s transactional DDL is impacted by unresolved issues in the pysqlite driver, which fails to emit BEGIN and additionally forces a COMMIT to cancel any transaction when DDL is encountered. See the section Serializable isolation / Savepoints / Transactional DDL for techniques to work around this behavior.

## Foreign Key Support

SQLite supports FOREIGN KEY syntax when emitting CREATE statements for tables, however by default these constraints have no effect on the operation of the table.

Constraint checking on SQLite has three prerequisites:

- At least version 3.6.19 of SQLite must be in use

- The SQLite library must be compiled without the SQLITE_OMIT_FOREIGN_KEY or SQLITE_OMIT_TRIGGER symbols enabled.

- The `PRAGMA foreign_keys = ON` statement must be emitted on all connections before use – including the initial call to `MetaData.create_all()`.

SQLAlchemy allows for the `PRAGMA` statement to be emitted automatically for new connections through the usage of events:

    from sqlalchemy.engine import Engine
    from sqlalchemy import event

    @event.listens_for(Engine, "connect")
    def set_sqlite_pragma(dbapi_connection, connection_record):
        cursor = dbapi_connection.cursor()
        cursor.execute("PRAGMA foreign_keys=ON")
        cursor.close()

Warning

When SQLite foreign keys are enabled, it is not possible to emit CREATE or DROP statements for tables that contain mutually-dependent foreign key constraints; to emit the DDL for these tables requires that ALTER TABLE be used to create or drop these constraints separately, for which SQLite has no support.

See also

SQLite Foreign Key Support - on the SQLite web site.

Events - SQLAlchemy event API.

Creating/Dropping Foreign Key Constraints via ALTER - more information on SQLAlchemy’s facilities for handling

  mutually-dependent foreign key constraints.

## ON CONFLICT support for constraints

See also

This section describes the DDL version of “ON CONFLICT” for SQLite, which occurs within a CREATE TABLE statement. For “ON CONFLICT” as applied to an INSERT statement, see INSERT…ON CONFLICT (Upsert).

SQLite supports a non-standard DDL clause known as ON CONFLICT which can be applied to primary key, unique, check, and not null constraints. In DDL, it is rendered either within the “CONSTRAINT” clause or within the column definition itself depending on the location of the target constraint. To render this clause within DDL, the extension parameter `sqlite_on_conflict` can be specified with a string conflict resolution algorithm within the `PrimaryKeyConstraint`, `UniqueConstraint`, `CheckConstraint` objects. Within the `Column` object, there are individual parameters `sqlite_on_conflict_not_null`, `sqlite_on_conflict_primary_key`, `sqlite_on_conflict_unique` which each correspond to the three types of relevant constraint types that can be indicated from a `Column` object.

See also

ON CONFLICT - in the SQLite documentation

New in version 1.3.

The `sqlite_on_conflict` parameters accept a string argument which is just the resolution name to be chosen, which on SQLite can be one of ROLLBACK, ABORT, FAIL, IGNORE, and REPLACE. For example, to add a UNIQUE constraint that specifies the IGNORE algorithm:

    some_table = Table(
        'some_table', metadata,
        Column('id', Integer, primary_key=True),
        Column('data', Integer),
        UniqueConstraint('id', 'data', sqlite_on_conflict='IGNORE')
    )

The above renders CREATE TABLE DDL as:

    CREATE TABLE some_table (
        id INTEGER NOT NULL,
        data INTEGER,
        PRIMARY KEY (id),
        UNIQUE (id, data) ON CONFLICT IGNORE
    )

When using the `Column.unique` flag to add a UNIQUE constraint to a single column, the `sqlite_on_conflict_unique` parameter can be added to the `Column` as well, which will be added to the UNIQUE constraint in the DDL:

    some_table = Table(
        'some_table', metadata,
        Column('id', Integer, primary_key=True),
        Column('data', Integer, unique=True,
               sqlite_on_conflict_unique='IGNORE')
    )

rendering:

    CREATE TABLE some_table (
        id INTEGER NOT NULL,
        data INTEGER,
        PRIMARY KEY (id),
        UNIQUE (data) ON CONFLICT IGNORE
    )

To apply the FAIL algorithm for a NOT NULL constraint, `sqlite_on_conflict_not_null` is used:

    some_table = Table(
        'some_table', metadata,
        Column('id', Integer, primary_key=True),
        Column('data', Integer, nullable=False,
               sqlite_on_conflict_not_null='FAIL')
    )

this renders the column inline ON CONFLICT phrase:

    CREATE TABLE some_table (
        id INTEGER NOT NULL,
        data INTEGER NOT NULL ON CONFLICT FAIL,
        PRIMARY KEY (id)
    )

Similarly, for an inline primary key, use `sqlite_on_conflict_primary_key`:

    some_table = Table(
        'some_table', metadata,
        Column('id', Integer, primary_key=True,
               sqlite_on_conflict_primary_key='FAIL')
    )

SQLAlchemy renders the PRIMARY KEY constraint separately, so the conflict resolution algorithm is applied to the constraint itself:

    CREATE TABLE some_table (
        id INTEGER NOT NULL,
        PRIMARY KEY (id) ON CONFLICT FAIL
    )

## INSERT…ON CONFLICT (Upsert)

See also

This section describes the DML version of “ON CONFLICT” for SQLite, which occurs within an INSERT statement. For “ON CONFLICT” as applied to a CREATE TABLE statement, see ON CONFLICT support for constraints.

From version 3.24.0 onwards, SQLite supports “upserts” (update or insert) of rows into a table via the `ON CONFLICT` clause of the `INSERT` statement. A candidate row will only be inserted if that row does not violate any unique or primary key constraints. In the case of a unique constraint violation, a secondary action can occur which can be either “DO UPDATE”, indicating that the data in the target row should be updated, or “DO NOTHING”, which indicates to silently skip this row.

Conflicts are determined using columns that are part of existing unique constraints and indexes. These constraints are identified by stating the columns and conditions that comprise the indexes.

SQLAlchemy provides `ON CONFLICT` support via the SQLite-specific `insert()` function, which provides the generative methods `Insert.on_conflict_do_update()` and `Insert.on_conflict_do_nothing()`:

    >>> from sqlalchemy.dialects.sqlite import insert

    >>> insert_stmt = insert(my_table).values(
    ...     id='some_existing_id',
    ...     data='inserted value')

    >>> do_update_stmt = insert_stmt.on_conflict_do_update(
    ...     index_elements=['id'],
    ...     set_=dict(data='updated value')
    ... )

    >>> print(do_update_stmt)
    INSERT INTO my_table (id, data) VALUES (?, ?)
    ON CONFLICT (id) DO UPDATE SET data = ?
    >>> do_nothing_stmt = insert_stmt.on_conflict_do_nothing(
    ...     index_elements=['id']
    ... )

    >>> print(do_nothing_stmt)
    INSERT INTO my_table (id, data) VALUES (?, ?)
    ON CONFLICT (id) DO NOTHING

New in version 1.4.

See also

Upsert - in the SQLite documentation.

### Specifying the Target

Both methods supply the “target” of the conflict using column inference:

- The `Insert.on_conflict_do_update.index_elements` argument specifies a sequence containing string column names, `Column` objects, and/or SQL expression elements, which would identify a unique index or unique constraint.

- When using `Insert.on_conflict_do_update.index_elements` to infer an index, a partial index can be inferred by also specifying the `Insert.on_conflict_do_update.index_where` parameter:

      >>> stmt = insert(my_table).values(user_email='a@b.com', data='inserted data')

      >>> do_update_stmt = stmt.on_conflict_do_update(
      ...     index_elements=[my_table.c.user_email],
      ...     index_where=my_table.c.user_email.like('%@gmail.com'),
      ...     set_=dict(data=stmt.excluded.data)
      ...     )

      >>> print(do_update_stmt)
      INSERT INTO my_table (data, user_email) VALUES (?, ?)
      ON CONFLICT (user_email)
      WHERE user_email LIKE '%@gmail.com'
      DO UPDATE SET data = excluded.data
      >>>

### The SET Clause

`ON CONFLICT...DO UPDATE` is used to perform an update of the already existing row, using any combination of new values as well as values from the proposed insertion. These values are specified using the `Insert.on_conflict_do_update.set_` parameter. This parameter accepts a dictionary which consists of direct values for UPDATE:

    >>> stmt = insert(my_table).values(id='some_id', data='inserted value')

    >>> do_update_stmt = stmt.on_conflict_do_update(
    ...     index_elements=['id'],
    ...     set_=dict(data='updated value')
    ... )

    >>> print(do_update_stmt)

    INSERT INTO my_table (id, data) VALUES (?, ?)
    ON CONFLICT (id) DO UPDATE SET data = ?

Warning

The `Insert.on_conflict_do_update()` method does not take into account Python-side default UPDATE values or generation functions, e.g. those specified using `Column.onupdate`. These values will not be exercised for an ON CONFLICT style of UPDATE, unless they are manually specified in the `Insert.on_conflict_do_update.set_` dictionary.

### Updating using the Excluded INSERT Values

In order to refer to the proposed insertion row, the special alias `Insert.excluded` is available as an attribute on the `Insert` object; this object creates an “excluded.” prefix on a column, that informs the DO UPDATE to update the row with the value that would have been inserted had the constraint not failed:

    >>> stmt = insert(my_table).values(
    ...     id='some_id',
    ...     data='inserted value',
    ...     author='jlh'
    ... )

    >>> do_update_stmt = stmt.on_conflict_do_update(
    ...     index_elements=['id'],
    ...     set_=dict(data='updated value', author=stmt.excluded.author)
    ... )

    >>> print(do_update_stmt)
    INSERT INTO my_table (id, data, author) VALUES (?, ?, ?)
    ON CONFLICT (id) DO UPDATE SET data = ?, author = excluded.author

### Additional WHERE Criteria

The `Insert.on_conflict_do_update()` method also accepts a WHERE clause using the `Insert.on_conflict_do_update.where` parameter, which will limit those rows which receive an UPDATE:

    >>> stmt = insert(my_table).values(
    ...     id='some_id',
    ...     data='inserted value',
    ...     author='jlh'
    ... )

    >>> on_update_stmt = stmt.on_conflict_do_update(
    ...     index_elements=['id'],
    ...     set_=dict(data='updated value', author=stmt.excluded.author),
    ...     where=(my_table.c.status == 2)
    ... )
    >>> print(on_update_stmt)
    INSERT INTO my_table (id, data, author) VALUES (?, ?, ?)
    ON CONFLICT (id) DO UPDATE SET data = ?, author = excluded.author
    WHERE my_table.status = ?

### Skipping Rows with DO NOTHING

`ON CONFLICT` may be used to skip inserting a row entirely if any conflict with a unique constraint occurs; below this is illustrated using the `Insert.on_conflict_do_nothing()` method:

    >>> stmt = insert(my_table).values(id='some_id', data='inserted value')
    >>> stmt = stmt.on_conflict_do_nothing(index_elements=['id'])
    >>> print(stmt)
    INSERT INTO my_table (id, data) VALUES (?, ?) ON CONFLICT (id) DO NOTHING

If `DO NOTHING` is used without specifying any columns or constraint, it has the effect of skipping the INSERT for any unique violation which occurs:

    >>> stmt = insert(my_table).values(id='some_id', data='inserted value')
    >>> stmt = stmt.on_conflict_do_nothing()
    >>> print(stmt)
    INSERT INTO my_table (id, data) VALUES (?, ?) ON CONFLICT DO NOTHING

## Type Reflection

SQLite types are unlike those of most other database backends, in that the string name of the type usually does not correspond to a “type” in a one-to-one fashion. Instead, SQLite links per-column typing behavior to one of five so-called “type affinities” based on a string matching pattern for the type.

SQLAlchemy’s reflection process, when inspecting types, uses a simple lookup table to link the keywords returned to provided SQLAlchemy types. This lookup table is present within the SQLite dialect as it is for all other dialects. However, the SQLite dialect has a different “fallback” routine for when a particular type name is not located in the lookup map; it instead implements the SQLite “type affinity” scheme located at https://www.sqlite.org/datatype3.html section 2.1.

The provided typemap will make direct associations from an exact string name match for the following types:

`BIGINT`, `BLOB`, `BOOLEAN`, `BOOLEAN`, `CHAR`, `DATE`, `DATETIME`, `FLOAT`, `DECIMAL`, `FLOAT`, `INTEGER`, `INTEGER`, `NUMERIC`, `REAL`, `SMALLINT`, `TEXT`, `TIME`, `TIMESTAMP`, `VARCHAR`, `NVARCHAR`, `NCHAR`

When a type name does not match one of the above types, the “type affinity” lookup is used instead:

- `INTEGER` is returned if the type name includes the string `INT`

- `TEXT` is returned if the type name includes the string `CHAR`, `CLOB` or `TEXT`

- `NullType` is returned if the type name includes the string `BLOB`

- `REAL` is returned if the type name includes the string `REAL`, `FLOA` or `DOUB`.

- Otherwise, the `NUMERIC` type is used.

New in version 0.9.3: Support for SQLite type affinity rules when reflecting columns.

## Partial Indexes

A partial index, e.g. one which uses a WHERE clause, can be specified with the DDL system using the argument `sqlite_where`:

    tbl = Table('testtbl', m, Column('data', Integer))
    idx = Index('test_idx1', tbl.c.data,
                sqlite_where=and_(tbl.c.data > 5, tbl.c.data < 10))

The index will be rendered at create time as:

    CREATE INDEX test_idx1 ON testtbl (data)
    WHERE data > 5 AND data < 10

New in version 0.9.9.

## Dotted Column Names

Using table or column names that explicitly have periods in them is not recommended. While this is generally a bad idea for relational databases in general, as the dot is a syntactically significant character, the SQLite driver up until version 3.10.0 of SQLite has a bug which requires that SQLAlchemy filter out these dots in result sets.

Changed in version 1.1: The following SQLite issue has been resolved as of version 3.10.0 of SQLite. SQLAlchemy as of 1.1 automatically disables its internal workarounds based on detection of this version.

The bug, entirely outside of SQLAlchemy, can be illustrated thusly:

    import sqlite3

    assert sqlite3.sqlite_version_info < (3, 10, 0), "bug is fixed in this version"

    conn = sqlite3.connect(":memory:")
    cursor = conn.cursor()

    cursor.execute("create table x (a integer, b integer)")
    cursor.execute("insert into x (a, b) values (1, 1)")
    cursor.execute("insert into x (a, b) values (2, 2)")

    cursor.execute("select x.a, x.b from x")
    assert [c[0] for c in cursor.description] == ['a', 'b']

    cursor.execute('''
        select x.a, x.b from x where a=1
        union
        select x.a, x.b from x where a=2
    ''')
    assert [c[0] for c in cursor.description] == ['a', 'b'], \
        [c[0] for c in cursor.description]

The second assertion fails:

    Traceback (most recent call last):
      File "test.py", line 19, in <module>
        [c[0] for c in cursor.description]
    AssertionError: ['x.a', 'x.b']

Where above, the driver incorrectly reports the names of the columns including the name of the table, which is entirely inconsistent vs. when the UNION is not present.

SQLAlchemy relies upon column names being predictable in how they match to the original statement, so the SQLAlchemy dialect has no choice but to filter these out:

    from sqlalchemy import create_engine

    eng = create_engine("sqlite://")
    conn = eng.connect()

    conn.exec_driver_sql("create table x (a integer, b integer)")
    conn.exec_driver_sql("insert into x (a, b) values (1, 1)")
    conn.exec_driver_sql("insert into x (a, b) values (2, 2)")

    result = conn.exec_driver_sql("select x.a, x.b from x")
    assert result.keys() == ["a", "b"]

    result = conn.exec_driver_sql('''
        select x.a, x.b from x where a=1
        union
        select x.a, x.b from x where a=2
    ''')
    assert result.keys() == ["a", "b"]

Note that above, even though SQLAlchemy filters out the dots, both names are still addressable:

    >>> row = result.first()
    >>> row["a"]
    1
    >>> row["x.a"]
    1
    >>> row["b"]
    1
    >>> row["x.b"]
    1

Therefore, the workaround applied by SQLAlchemy only impacts `CursorResult.keys()` and `Row.keys()` in the public API. In the very specific case where an application is forced to use column names that contain dots, and the functionality of `CursorResult.keys()` and `Row.keys()` is required to return these dotted names unmodified, the `sqlite_raw_colnames` execution option may be provided, either on a per-`Connection` basis:

    result = conn.execution_options(sqlite_raw_colnames=True).exec_driver_sql('''
        select x.a, x.b from x where a=1
        union
        select x.a, x.b from x where a=2
    ''')
    assert result.keys() == ["x.a", "x.b"]

or on a per-`Engine` basis:

    engine = create_engine("sqlite://", execution_options={"sqlite_raw_colnames": True})

When using the per-`Engine` execution option, note that Core and ORM queries that use UNION may not function properly.

## SQLite-specific table options

One option for CREATE TABLE is supported directly by the SQLite dialect in conjunction with the `Table` construct:

- `WITHOUT ROWID`:

      Table("some_table", metadata, ..., sqlite_with_rowid=False)

See also

SQLite CREATE TABLE options

## SQLite Data Types

As with all SQLAlchemy dialects, all UPPERCASE types that are known to be valid with SQLite are importable from the top level dialect, whether they originate from `sqlalchemy.types` or from the local dialect:

    from sqlalchemy.dialects.sqlite import (
        BLOB,
        BOOLEAN,
        CHAR,
        DATE,
        DATETIME,
        DECIMAL,
        FLOAT,
        INTEGER,
        NUMERIC,
        JSON,
        SMALLINT,
        TEXT,
        TIME,
        TIMESTAMP,
        VARCHAR,
    )

Object Name | Description
--- | ---
DATE | Represent a Python date object in SQLite using a string.
DATETIME | Represent a Python datetime object in SQLite using a string.
JSON | SQLite JSON type.
TIME | Represent a Python time object in SQLite using a string.

class sqlalchemy.dialects.sqlite.DATETIME(*args, **kwargs)

  Represent a Python datetime object in SQLite using a string.

  The default string storage format is:

      "%(year)04d-%(month)02d-%(day)02d %(hour)02d:%(minute)02d:%(second)02d.%(microsecond)06d"

  e.g.:

      2021-03-15 12:05:57.105542

  The storage format can be customized to some degree using the `storage_format` and `regexp` parameters, such as:

      import re
      from sqlalchemy.dialects.sqlite import DATETIME

      dt = DATETIME(storage_format="%(year)04d/%(month)02d/%(day)02d "
                                   "%(hour)02d:%(minute)02d:%(second)02d",
                    regexp=r"(\d+)/(\d+)/(\d+) (\d+)-(\d+)-(\d+)"
      )

  Parameters:

    - storage_format – format string which will be applied to the dict with keys year, month, day, hour, minute, second, and microsecond.

    - regexp – regular expression which will be applied to incoming result rows. If the regexp contains named groups, the resulting match dict is applied to the Python datetime() constructor as keyword arguments. Otherwise, if positional groups are used, the datetime() constructor is called with positional arguments via `*map(int, match_obj.groups(0))`.

  Class signature

  class `sqlalchemy.dialects.sqlite.DATETIME` (`sqlalchemy.dialects.sqlite.base._DateTimeMixin`, `sqlalchemy.types.DateTime`)

class sqlalchemy.dialects.sqlite.DATE(storage_format=None, regexp=None, **kw)

  Represent a Python date object in SQLite using a string.

  The default string storage format is:

      "%(year)04d-%(month)02d-%(day)02d"

  e.g.:

      2011-03-15

  The storage format can be customized to some degree using the `storage_format` and `regexp` parameters, such as:

      import re
      from sqlalchemy.dialects.sqlite import DATE

      d = DATE(
              storage_format="%(month)02d/%(day)02d/%(year)04d",
              regexp=re.compile("(?P<month>\d+)/(?P<day>\d+)/(?P<year>\d+)")
          )

  Parameters:

    - storage_format – format string which will be applied to the dict with keys year, month, and day.

    - regexp – regular expression which will be applied to incoming result rows. If the regexp contains named groups, the resulting match dict is applied to the Python date() constructor as keyword arguments. Otherwise, if positional groups are used, the date() constructor is called with positional arguments via `*map(int, match_obj.groups(0))`.

  Class signature

  class `sqlalchemy.dialects.sqlite.DATE` (`sqlalchemy.dialects.sqlite.base._DateTimeMixin`, `sqlalchemy.types.Date`)

class sqlalchemy.dialects.sqlite.JSON(none_as_null=False)

  SQLite JSON type.

  SQLite supports JSON as of version 3.9 through its JSON1 extension. Note that JSON1 is a loadable extension and as such may not be available, or may require run-time loading.

  `JSON` is used automatically whenever the base `JSON` datatype is used against a SQLite backend.

  See also

  `JSON` - main documentation for the generic cross-platform JSON datatype.

  The `JSON` type supports persistence of JSON values as well as the core index operations provided by `JSON` datatype, by adapting the operations to render the `JSON_EXTRACT` function wrapped in the `JSON_QUOTE` function at the database level. Extracted values are quoted in order to ensure that the results are always JSON string values.

  New in version 1.3.

  Class signature

  class `sqlalchemy.dialects.sqlite.JSON` (`sqlalchemy.types.JSON`)

class sqlalchemy.dialects.sqlite.TIME(*args, **kwargs)

  Represent a Python time object in SQLite using a string.

  The default string storage format is:

      "%(hour)02d:%(minute)02d:%(second)02d.%(microsecond)06d"

  e.g.:

      12:05:57.10558

  The storage format can be customized to some degree using the `storage_format` and `regexp` parameters, such as:

      import re
      from sqlalchemy.dialects.sqlite import TIME

      t = TIME(storage_format="%(hour)02d-%(minute)02d-"
                              "%(second)02d-%(microsecond)06d",
               regexp=re.compile("(\d+)-(\d+)-(\d+)-(?:-(\d+))?")
      )

  Parameters:

    - storage_format – format string which will be applied to the dict with keys hour, minute, second, and microsecond.

    - regexp – regular expression which will be applied to incoming result rows. If the regexp contains named groups, the resulting match dict is applied to the Python time() constructor as keyword arguments. Otherwise, if positional groups are used, the time() constructor is called with positional arguments via `*map(int, match_obj.groups(0))`.

  Class signature

  class `sqlalchemy.dialects.sqlite.TIME` (`sqlalchemy.dialects.sqlite.base._DateTimeMixin`, `sqlalchemy.types.Time`)

## SQLite DML Constructs

Object Name | Description
--- | ---
insert(table[, values, inline, bind, ...], **dialect_kw) | Construct an `Insert` object.
Insert | SQLite-specific implementation of INSERT.

function sqlalchemy.dialects.sqlite.insert(table, values=None, inline=False, bind=None, prefixes=None, returning=None, return_defaults=False, **dialect_kw)

  Construct an `Insert` object.

  This documentation is inherited from `sqlalchemy.sql.expression.insert()`; this constructor, `sqlalchemy.dialects.sqlite.insert()`, creates a `sqlalchemy.dialects.sqlite.Insert` object. See that class for additional details describing this subclass.

  E.g.:

      from sqlalchemy import insert

      stmt = (
          insert(user_table).
          values(name='username', fullname='Full Username')
      )

  Similar functionality is available via the `TableClause.insert()` method on `Table`.

  See also

  Inserting Rows with Core - in the SQLAlchemy 1.4 / 2.0 Tutorial

  Parameters:

    - table – `TableClause` which is the subject of the insert.

    - values –

      collection of values to be inserted; see `Insert.values()` for a description of allowed formats here. Can be omitted entirely; a `Insert` construct will also dynamically render the VALUES clause at execution time based on the parameters passed to `Connection.execute()`.

      Deprecated since version 1.4: The `insert.values` parameter will be removed in SQLAlchemy 2.0. Please refer to the `Insert.values()` method.

    - inline –

      if True, no attempt will be made to retrieve the SQL-generated default values to be provided within the statement; in particular, this allows SQL expressions to be rendered ‘inline’ within the statement without the need to pre-execute them beforehand; for backends that support “returning”, this turns off the “implicit returning” feature for the statement.

      Deprecated since version 1.4: The `insert.inline` parameter will be removed in SQLAlchemy 2.0. Please use the `Insert.inline()` method.

  If both `Insert.values` and compile-time bind parameters are present, the compile-time bind parameters override the information specified within `Insert.values` on a per-key basis.

  The keys within `Insert.values` can be either `Column` objects or their string identifiers. Each key may reference one of:

  - a literal data value (i.e. string, number, etc.);

  - a Column object;

  - a SELECT statement.

  If a `SELECT` statement is specified which references this `INSERT` statement’s table, the statement will be correlated against the `INSERT` statement.

  See also

  Inserting Rows with Core - in the SQLAlchemy 1.4 / 2.0 Tutorial

class sqlalchemy.dialects.sqlite.Insert(table, values=None, inline=False, bind=None, prefixes=None, returning=None, return_defaults=False, **dialect_kw)

  SQLite-specific implementation of INSERT.

  Adds methods for SQLite-specific syntaxes such as ON CONFLICT.

  The `Insert` object is created using the `sqlalchemy.dialects.sqlite.insert()` function.

  New in version 1.4.

  See also

  INSERT…ON CONFLICT (Upsert)

  Members

  excluded, inherit_cache, on_conflict_do_nothing(), on_conflict_do_update()

  Class signature

  class `sqlalchemy.dialects.sqlite.Insert` (`sqlalchemy.sql.expression.Insert`)

  attribute `sqlalchemy.dialects.sqlite.Insert.` excluded

    Provide the `excluded` namespace for an ON CONFLICT statement

    SQLite’s ON CONFLICT clause allows reference to the row that would be inserted, known as `excluded`. This attribute provides all columns in this row to be referenceable.

    Tip

    The `Insert.excluded` attribute is an instance of `ColumnCollection`, which provides an interface the same as that of the `Table.c` collection described at Accessing Tables and Columns. With this collection, ordinary names are accessible like attributes (e.g. `stmt.excluded.some_column`), but special names and dictionary method names should be accessed using indexed access, such as `stmt.excluded["column name"]` or `stmt.excluded["values"]`. See the docstring for `ColumnCollection` for further examples.

  attribute `sqlalchemy.dialects.sqlite.Insert.` inherit_cache = False

    Indicate if this `HasCacheKey` instance should make use of the cache key generation scheme used by its immediate superclass.

    The attribute defaults to `None`, which indicates that a construct has not yet taken into account whether or not its appropriate for it to participate in caching; this is functionally equivalent to setting the value to `False`, except that a warning is also emitted.

    This flag can be set to `True` on a particular class, if the SQL that corresponds to the object does not change based on attributes which are local to this class, and not its superclass.

    See also

    Enabling Caching Support for Custom Constructs - General guideslines for setting the `HasCacheKey.inherit_cache` attribute for third-party or user defined SQL constructs.

  method `sqlalchemy.dialects.sqlite.Insert.` on_conflict_do_nothing(index_elements=None, index_where=None)

    Specifies a DO NOTHING action for ON CONFLICT clause.

    Parameters:

      - index_elements – A sequence consisting of string column names, `Column` objects, or other column expression objects that will be used to infer a target index or unique constraint.

      - index_where – Additional WHERE criterion that can be used to infer a conditional target index.

  method `sqlalchemy.dialects.sqlite.Insert.` on_conflict_do_update(index_elements=None, index_where=None, set_=None, where=None)

    Specifies a DO UPDATE SET action for ON CONFLICT clause.

    Parameters:

      - index_elements – A sequence consisting of string column names, `Column` objects, or other column expression objects that will be used to infer a target index or unique constraint.

      - index_where – Additional WHERE criterion that can be used to infer a conditional target index.

      - set_ –

        A dictionary or other mapping object where the keys are either names of columns in the target table, or `Column` objects or other ORM-mapped columns matching that of the target table, and expressions or literals as values, specifying the `SET` actions to take.

        New in version 1.4: The `Insert.on_conflict_do_update.set_` parameter supports `Column` objects from the target `Table` as keys.

        Warning

        This dictionary does not take into account Python-specified default UPDATE values or generation functions, e.g. those specified using `Column.onupdate`. These values will not be exercised for an ON CONFLICT style of UPDATE, unless they are manually specified in the `Insert.on_conflict_do_update.set_` dictionary.

      - where – Optional argument. If present, can be a literal SQL string or an acceptable expression for a `WHERE` clause that restricts the rows affected by `DO UPDATE SET`. Rows not meeting the `WHERE` condition will not be updated (effectively a `DO NOTHING` for those rows).

## Pysqlite

Support for the SQLite database via the pysqlite driver.

Note that `pysqlite` is the same driver as the `sqlite3` module included with the Python distribution.

### DBAPI

Documentation and download information (if applicable) for pysqlite is available at: https://docs.python.org/library/sqlite3.html

### Connecting

Connect String:

    sqlite+pysqlite:///file_path

### Driver

The `sqlite3` Python DBAPI is standard on all modern Python versions; for cPython and Pypy, no additional installation is necessary.

### Connect Strings

The file specification for the SQLite database is taken as the “database” portion of the URL. Note that the format of a SQLAlchemy url is:

    driver://user:pass@host/database

This means that the actual filename to be used starts with the characters to the right of the third slash. So connecting to a relative filepath looks like:

    # relative path
    e = create_engine('sqlite:///path/to/database.db')

An absolute path, which is denoted by starting with a slash, means you need four slashes:

    # absolute path
    e = create_engine('sqlite:////path/to/database.db')

To use a Windows path, regular drive specifications and backslashes can be used. Double backslashes are probably needed:

    # absolute path on Windows
    e = create_engine('sqlite:///C:\\path\\to\\database.db')

The sqlite `:memory:` identifier is the default if no filepath is present. Specify `sqlite://` and nothing else:

    # in-memory database
    e = create_engine('sqlite://')

#### URI Connections

Modern versions of SQLite support an alternative system of connecting using a driver level URI, which has the advantage that additional driver-level arguments can be passed including options such as “read only”. The Python sqlite3 driver supports this mode under modern Python 3 versions. The SQLAlchemy pysqlite driver supports this mode of use by specifying “uri=true” in the URL query string. The SQLite-level “URI” is kept as the “database” portion of the SQLAlchemy url (that is, following a slash):

    e = create_engine("sqlite:///file:path/to/database?mode=ro&uri=true")

Note

The “uri=true” parameter must appear in the query string of the URL. It will not currently work as expected if it is only present in the `create_engine.connect_args` parameter dictionary.

The logic reconciles the simultaneous presence of SQLAlchemy’s query string and SQLite’s query string by separating out the parameters that belong to the Python sqlite3 driver vs. those that belong to the SQLite URI. This is achieved through the use of a fixed list of parameters known to be accepted by the Python side of the driver. For example, to include a URL that indicates the Python sqlite3 “timeout” and “check_same_thread” parameters, along with the SQLite “mode” and “nolock” parameters, they can all be passed together on the query string:

    e = create_engine(
        "sqlite:///file:path/to/database?"
        "check_same_thread=true&timeout=10&mode=ro&nolock=1&uri=true"
    )

Above, the pysqlite / sqlite3 DBAPI would be passed arguments as:

    sqlite3.connect(
        "file:path/to/database?mode=ro&nolock=1",
        check_same_thread=True, timeout=10, uri=True
    )

Regarding future parameters added to either the Python or native drivers. new parameter names added to the SQLite URI scheme should be automatically accommodated by this scheme. New parameter names added to the Python driver side can be accommodated by specifying them in the `create_engine.connect_args` dictionary, until dialect support is added by SQLAlchemy. For the less likely case that the native SQLite driver adds a new parameter name that overlaps with one of the existing, known Python driver parameters (such as “timeout” perhaps), SQLAlchemy’s dialect would require adjustment for the URL scheme to continue to support this.

As is always the case for all SQLAlchemy dialects, the entire “URL” process can be bypassed in `create_engine()` through the use of the `create_engine.creator` parameter which allows for a custom callable that creates a Python sqlite3 driver level connection directly.

New in version 1.3.9.

See also

Uniform Resource Identifiers - in the SQLite documentation

### Regular Expression Support

New in version 1.4.

Support for the `ColumnOperators.regexp_match()` operator is provided using Python’s re.search function. SQLite itself does not include a working regular expression operator; instead, it includes a non-implemented placeholder operator `REGEXP` that calls a user-defined function that must be provided.

SQLAlchemy’s implementation makes use of the pysqlite create_function hook as follows:

    def regexp(a, b):
        return re.search(a, b) is not None

    sqlite_connection.create_function(
        "regexp", 2, regexp,
    )

There is currently no support for regular expression flags as a separate argument, as these are not supported by SQLite’s REGEXP operator, however these may be included inline within the regular expression string. See Python regular expressions for details.

See also

Python regular expressions: Documentation for Python’s regular expression syntax.

### Compatibility with sqlite3 “native” date and datetime types

The pysqlite driver includes the sqlite3.PARSE_DECLTYPES and sqlite3.PARSE_COLNAMES options, which have the effect of any column or expression explicitly cast as “date” or “timestamp” will be converted to a Python date or datetime object. The date and datetime types provided with the pysqlite dialect are not currently compatible with these options, since they render the ISO date/datetime including microseconds, which pysqlite’s driver does not. Additionally, SQLAlchemy does not at this time automatically render the “cast” syntax required for the freestanding functions “current_timestamp” and “current_date” to return datetime/date types natively. Unfortunately, pysqlite does not provide the standard DBAPI types in `cursor.description`, leaving SQLAlchemy with no way to detect these types on the fly without expensive per-row type checks.

Keeping in mind that pysqlite’s parsing option is not recommended, nor should be necessary, for use with SQLAlchemy, usage of PARSE_DECLTYPES can be forced if one configures “native_datetime=True” on create_engine():

    engine = create_engine('sqlite://',
        connect_args={'detect_types':
            sqlite3.PARSE_DECLTYPES|sqlite3.PARSE_COLNAMES},
        native_datetime=True
    )

With this flag enabled, the DATE and TIMESTAMP types (but note - not the DATETIME or TIME types…confused yet ?) will not perform any bind parameter or result processing. Execution of “func.current_date()” will return a string. “func.current_timestamp()” is registered as returning a DATETIME type in SQLAlchemy, so this function still receives SQLAlchemy-level result processing.

### Threading/Pooling Behavior

Pysqlite’s default behavior is to prohibit the usage of a single connection in more than one thread. This is originally intended to work with older versions of SQLite that did not support multithreaded operation under various circumstances. In particular, older SQLite versions did not allow a `:memory:` database to be used in multiple threads under any circumstances.

Pysqlite does include a now-undocumented flag known as `check_same_thread` which will disable this check, however note that pysqlite connections are still not safe to use in concurrently in multiple threads. In particular, any statement execution calls would need to be externally mutexed, as Pysqlite does not provide for thread-safe propagation of error messages among other things. So while even `:memory:` databases can be shared among threads in modern SQLite, Pysqlite doesn’t provide enough thread-safety to make this usage worth it.

SQLAlchemy sets up pooling to work with Pysqlite’s default behavior:

- When a `:memory:` SQLite database is specified, the dialect by default will use `SingletonThreadPool`. This pool maintains a single connection per thread, so that all access to the engine within the current thread use the same `:memory:` database - other threads would access a different `:memory:` database.

- When a file-based database is specified, the dialect will use `NullPool` as the source of connections. This pool closes and discards connections which are returned to the pool immediately. SQLite file-based connections have extremely low overhead, so pooling is not necessary. The scheme also prevents a connection from being used again in a different thread and works best with SQLite’s coarse-grained file locking.

#### Using a Memory Database in Multiple Threads

To use a `:memory:` database in a multithreaded scenario, the same connection object must be shared among threads, since the database exists only within the scope of that connection. The `StaticPool` implementation will maintain a single connection globally, and the `check_same_thread` flag can be passed to Pysqlite as `False`:

    from sqlalchemy.pool import StaticPool
    engine = create_engine('sqlite://',
                        connect_args={'check_same_thread':False},
                        poolclass=StaticPool)

Note that using a `:memory:` database in multiple threads requires a recent version of SQLite.

#### Using Temporary Tables with SQLite

Due to the way SQLite deals with temporary tables, if you wish to use a temporary table in a file-based SQLite database across multiple checkouts from the connection pool, such as when using an ORM `Session` where the temporary table should continue to remain after `Session.commit()` or `Session.rollback()` is called, a pool which maintains a single connection must be used. Use `SingletonThreadPool` if the scope is only needed within the current thread, or `StaticPool` is scope is needed within multiple threads for this case:

    # maintain the same connection per thread
    from sqlalchemy.pool import SingletonThreadPool
    engine = create_engine('sqlite:///mydb.db',
                        poolclass=SingletonThreadPool)

    # maintain the same connection across all threads
    from sqlalchemy.pool import StaticPool
    engine = create_engine('sqlite:///mydb.db',
                        poolclass=StaticPool)

Note that `SingletonThreadPool` should be configured for the number of threads that are to be used; beyond that number, connections will be closed out in a non deterministic way.

### Unicode

The pysqlite driver only returns Python `unicode` objects in result sets, never plain strings, and accommodates `unicode` objects within bound parameter values in all cases. Regardless of the SQLAlchemy string type in use, string-based result values will by Python `unicode` in Python 2. The `Unicode` type should still be used to indicate those columns that require unicode, however, so that non-`unicode` values passed inadvertently will emit a warning. Pysqlite will emit an error if a non-`unicode` string is passed containing non-ASCII characters.

### Dealing with Mixed String / Binary Columns in Python 3

The SQLite database is weakly typed, and as such it is possible when using binary values, which in Python 3 are represented as `b'some string'`, that a particular SQLite database can have data values within different rows where some of them will be returned as a `b''` value by the Pysqlite driver, and others will be returned as Python strings, e.g. `''` values. This situation is not known to occur if the SQLAlchemy `LargeBinary` datatype is used consistently, however if a particular SQLite database has data that was inserted using the Pysqlite driver directly, or when using the SQLAlchemy `String` type which was later changed to `LargeBinary`, the table will not be consistently readable because SQLAlchemy’s `LargeBinary` datatype does not handle strings so it has no way of “encoding” a value that is in string format.

To deal with a SQLite table that has mixed string / binary data in the same column, use a custom type that will check each row individually:

    # note this is Python 3 only

    from sqlalchemy import String
    from sqlalchemy import TypeDecorator

    class MixedBinary(TypeDecorator):
        impl = String
        cache_ok = True

        def process_result_value(self, value, dialect):
            if isinstance(value, str):
                value = bytes(value, 'utf-8')
            elif value is not None:
                value = bytes(value)

            return value

Then use the above `MixedBinary` datatype in the place where `LargeBinary` would normally be used.

### Serializable isolation / Savepoints / Transactional DDL

In the section Database Locking Behavior / Concurrency, we refer to the pysqlite driver’s assortment of issues that prevent several features of SQLite from working correctly. The pysqlite DBAPI driver has several long-standing bugs which impact the correctness of its transactional behavior. In its default mode of operation, SQLite features such as SERIALIZABLE isolation, transactional DDL, and SAVEPOINT support are non-functional, and in order to use these features, workarounds must be taken.

The issue is essentially that the driver attempts to second-guess the user’s intent, failing to start transactions and sometimes ending them prematurely, in an effort to minimize the SQLite databases’s file locking behavior, even though SQLite itself uses “shared” locks for read-only activities.

SQLAlchemy chooses to not alter this behavior by default, as it is the long-expected behavior of the pysqlite driver; if and when the pysqlite driver attempts to repair these issues, that will be more of a driver towards defaults for SQLAlchemy.

The good news is that with a few events, we can implement transactional support fully, by disabling pysqlite’s feature entirely and emitting BEGIN ourselves. This is achieved using two event listeners:

    from sqlalchemy import create_engine, event

    engine = create_engine("sqlite:///myfile.db")

    @event.listens_for(engine, "connect")
    def do_connect(dbapi_connection, connection_record):
        # disable pysqlite's emitting of the BEGIN statement entirely.
        # also stops it from emitting COMMIT before any DDL.
        dbapi_connection.isolation_level = None

    @event.listens_for(engine, "begin")
    def do_begin(conn):
        # emit our own BEGIN
        conn.exec_driver_sql("BEGIN")

Warning

When using the above recipe, it is advised to not use the `Connection.execution_options.isolation_level` setting on `Connection` and `create_engine()` with the SQLite driver, as this function necessarily will also alter the “.isolation_level” setting.

Above, we intercept a new pysqlite connection and disable any transactional integration. Then, at the point at which SQLAlchemy knows that transaction scope is to begin, we emit `"BEGIN"` ourselves.

When we take control of `"BEGIN"`, we can also control directly SQLite’s locking modes, introduced at BEGIN TRANSACTION, by adding the desired locking mode to our `"BEGIN"`:

    @event.listens_for(engine, "begin")
    def do_begin(conn):
        conn.exec_driver_sql("BEGIN EXCLUSIVE")

See also

BEGIN TRANSACTION - on the SQLite site

sqlite3 SELECT does not BEGIN a transaction - on the Python bug tracker

sqlite3 module breaks transactions and potentially corrupts data - on the Python bug tracker

### User-Defined Functions

pysqlite supports a create_function() method that allows us to create our own user-defined functions (UDFs) in Python and use them directly in SQLite queries. These functions are registered with a specific DBAPI Connection.

SQLAlchemy uses connection pooling with file-based SQLite databases, so we need to ensure that the UDF is attached to the connection when it is created. That is accomplished with an event listener:

    from sqlalchemy import create_engine
    from sqlalchemy import event
    from sqlalchemy import text

    def udf():
        return "udf-ok"

    engine = create_engine("sqlite:///./db_file")

    @event.listens_for(engine, "connect")
    def connect(conn, rec):
        conn.create_function("udf", 0, udf)

    for i in range(5):
        with engine.connect() as conn:
            print(conn.scalar(text("SELECT UDF()")))

## Aiosqlite

Support for the SQLite database via the aiosqlite driver.

### DBAPI

Documentation and download information (if applicable) for aiosqlite is available at: https://pypi.org/project/aiosqlite/

### Connecting

Connect String:

    sqlite+aiosqlite:///file_path

The aiosqlite dialect provides support for the SQLAlchemy asyncio interface running on top of pysqlite.

aiosqlite is a wrapper around pysqlite that uses a background thread for each connection. It does not actually use non-blocking IO, as SQLite databases are not socket-based. However it does provide a working asyncio interface that’s useful for testing and prototyping purposes.

Using a special asyncio mediation layer, the aiosqlite dialect is usable as the backend for the SQLAlchemy asyncio extension package.

This dialect should normally be used only with the `create_async_engine()` engine creation function:

    from sqlalchemy.ext.asyncio import create_async_engine
    engine = create_async_engine("sqlite+aiosqlite:///filename")

The URL passes through all arguments to the `pysqlite` driver, so all connection arguments are the same as they are for that of Pysqlite.

### User-Defined Functions

aiosqlite extends pysqlite to support async, so we can create our own user-defined functions (UDFs) in Python and use them directly in SQLite queries as described here: User-Defined Functions.

## Pysqlcipher

Support for the SQLite database via the pysqlcipher driver.

Dialect for support of DBAPIs that make use of the SQLCipher backend.

### Connecting

Connect String:

    sqlite+pysqlcipher://:passphrase@/file_path[?kdf_iter=<iter>]

### Driver

Current dialect selection logic is:

- If the `create_engine.module` parameter supplies a DBAPI module, that module is used.

- Otherwise for Python 3, choose https://pypi.org/project/sqlcipher3/

- If not available, fall back to https://pypi.org/project/pysqlcipher3/

- For Python 2, https://pypi.org/project/pysqlcipher/ is used.

Warning

The `pysqlcipher3` and `pysqlcipher` DBAPI drivers are no longer maintained; the `sqlcipher3` driver as of this writing appears to be current. For future compatibility, any pysqlcipher-compatible DBAPI may be used as follows:

    import sqlcipher_compatible_driver

    from sqlalchemy import create_engine

    e = create_engine(
        "sqlite+pysqlcipher://:password@/dbname.db",
        module=sqlcipher_compatible_driver
    )

These drivers make use of the SQLCipher engine. This system essentially introduces new PRAGMA commands to SQLite which allows the setting of a passphrase and other encryption parameters, allowing the database file to be encrypted.

### Connect Strings

The format of the connect string is in every way the same as that of the `pysqlite` driver, except that the “password” field is now accepted, which should contain a passphrase:

    e = create_engine('sqlite+pysqlcipher://:testing@/foo.db')

For an absolute file path, two leading slashes should be used for the database name:

    e = create_engine('sqlite+pysqlcipher://:testing@//path/to/foo.db')

A selection of additional encryption-related pragmas supported by SQLCipher as documented at https://www.zetetic.net/sqlcipher/sqlcipher-api/ can be passed in the query string, and will result in that PRAGMA being called for each new connection. Currently, `cipher`, `kdf_iter` `cipher_page_size` and `cipher_use_hmac` are supported:

    e = create_engine('sqlite+pysqlcipher://:testing@/foo.db?cipher=aes-256-cfb&kdf_iter=64000')

Warning

Previous versions of sqlalchemy did not take into consideration the encryption-related pragmas passed in the url string, that were silently ignored. This may cause errors when opening files saved by a previous sqlalchemy version if the encryption options do not match.

### Pooling Behavior

The driver makes a change to the default pool behavior of pysqlite as described in Threading/Pooling Behavior. The pysqlcipher driver has been observed to be significantly slower on connection than the pysqlite driver, most likely due to the encryption overhead, so the dialect here defaults to using the `SingletonThreadPool` implementation, instead of the `NullPool` pool used by pysqlite. As always, the pool implementation is entirely configurable using the `create_engine.poolclass` parameter; the `StaticPool` may be more feasible for single-threaded use, or `NullPool` may be used to prevent unencrypted connections from being held open for long periods of time, at the expense of slower startup time for new connections.